///////////////////////////////////////////////////////////////////////
// WMV plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// File Information dialog (part of InputFileWMV class)

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	#define _CRT_SECURE_NO_WARNINGS
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <vfw.h>
#include <stdio.h>

#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"

#include "WinFile.h"
#include "DataVector.h"
#include "ASFStreamNode.h"
#include "InputFileWMV.h"
#include "Languages.h"
#include "Resource.h"

extern "C" const VDXPluginInfo *const * VDXAPIENTRY VDGetPluginInfo();


char *InputFileWMV::time_to_string(double seconds)
{
	// There are 86400 seconds in a day,
	// about 32 million seconds in a year.
	// Since we are unlikely to encounter videos
	// of such an extreme duration, we can easily
	// get away with using longs for this math...

	const int buf_size = 32;
	char *ret = new char[buf_size];

	if (ret != NULL)
	{
		int hours, mins, mils;
		int secs = (int)seconds;

		hours = secs / 3600;
		secs  = secs % 3600;
		mins  = secs / 60;
		secs  = secs % 60;
		mils  = (int)(seconds * 1000.0 + 0.5);
		mils  = mils % 1000;

		if (hours > 0)
		{
			_snprintf(ret, buf_size, "%i:%02i:%02i.%03i", hours, mins, secs, mils);
		}
		else
		{
			_snprintf(ret, buf_size, "%i:%02i.%03i", mins, secs, mils);
		}

		ret[buf_size - 1] = '\0';
	}

	return ret;
}


char *InputFileWMV::map_language(const char *lang)
{
	if (lang != NULL)
	{
		int i;
		for (i = 0; ; ++i)
		{
			const char *map = language_map[i][0];
			if (map == NULL) break;

			if (_stricmp(lang, map) == 0)
			{
				int  len  = (int)strlen(language_map[i][1]) + sizeof(char);
				char *ret = new char[len];

				if (ret != NULL)
				{
					memcpy(ret, language_map[i][1], len);
					return ret;
				}
			}
		}
	}

	return NULL;
}


void InputFileWMV::UpdateDlgAudio(HWND hDlg, int index)
{
	ACMFORMATTAGDETAILSA aftd;
	char buf[64];

	char *s;
	int i;

	struct {
		unsigned __int64 ui64StreamBytes;
		unsigned long dwFrameMinSize;
		unsigned long dwFrameAvgSize;
		unsigned long dwFrameMaxSize;
	} pInfo = {
		0, 0xFFFFFFFF, 0, 0
	};

	const WAVEFORMATEX  *wfex;
	const ASFStreamNode *pasn;

	InputFileWMV *thisPtr =
		(InputFileWMV *)GetWindowLongPtrA(hDlg, DWLP_USER);

	sprintf(buf, "Stream %i of %i", index + 1, thisPtr->num_audio_streams);

	// Get a pointer to the desired stream
	for (i = 0; i < thisPtr->num_streams; ++i)
	{
		if (thisPtr->streams[i]->fccType == streamtypeAUDIO)
		{
			if (index == 0)
			{
				pasn = thisPtr->streams[i];
				break;
			}
			--index;
		}
	}

/**************************************************************************
 *      Stream x of x
 **************************************************************************/

	// index out of bounds or no audio?
	if (i == thisPtr->num_streams)
	{
		SetDlgItemTextA(hDlg, IDC_STREAMNUM, "");
		return;
	}

	SetDlgItemTextA(hDlg, IDC_STREAMNUM, buf);

/**************************************************************************
 *      Sampling rate, channels:
 **************************************************************************/

	wfex = (WAVEFORMATEX *)pasn->pFormat;

	i = _snprintf(buf, sizeof(buf), "%lu Hz, ", wfex->nSamplesPerSec);

	if (i >= 0 && i < sizeof(buf))
	{
		if (wfex->nChannels == 1)
		{
			_snprintf(buf + i, sizeof(buf) - i, "Mono");
		}
		else if (wfex->nChannels == 2)
		{
			_snprintf(buf + i, sizeof(buf) - i, "Stereo");
		}
		else
		{
			_snprintf(buf + i, sizeof(buf) - i, "%hu channels", wfex->nChannels);
		}
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_SAMPLINGRATE, buf);

/**************************************************************************
 *      Sample precision:
 **************************************************************************/

	if (wfex->wFormatTag == WAVE_FORMAT_PCM && wfex->wBitsPerSample)
	{
		_snprintf(buf, sizeof(buf), "%hu-bit", wfex->wBitsPerSample);
	}
	else
	{
		_snprintf(buf, sizeof(buf), "N/A");
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_PRECISION, buf);

/**************************************************************************
 *      ACM codec (tag):
 **************************************************************************/

	memset(&aftd, 0, sizeof(aftd));
	aftd.cbStruct = sizeof(aftd);
	aftd.dwFormatTag = wfex->wFormatTag;

	if (acmFormatTagDetailsA(NULL, &aftd,
		ACM_FORMATTAGDETAILSF_FORMATTAG) == MMSYSERR_NOERROR)
	{
		_snprintf(buf, sizeof(buf), "%s (0x%04X)", aftd.szFormatTag, wfex->wFormatTag);
	}
	else
	{
		_snprintf(buf, sizeof(buf), "Missing codec (0x%04X)", wfex->wFormatTag);
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_COMPRESSION, buf);
			
/**************************************************************************
 *      Block layout (preroll):
 **************************************************************************/

	_snprintf(buf, sizeof(buf),
		"%lu chunks (%0.2fs preroll)",
		pasn->dwPerceivedLength,
		(double)thisPtr->dwPreroll / 1000.0);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_LAYOUT, buf);

/**************************************************************************
 *      Length (duration):
 **************************************************************************/

	s = time_to_string(pasn->Duration);
	if (s != NULL)
	{
		_snprintf(buf, sizeof(buf), "%lu samples (%s)",
			(unsigned long)(pasn->Duration * wfex->nSamplesPerSec + 0.5), s);

		delete[] s;
	}
	else
	{
		_snprintf(buf, sizeof(buf), "%lu samples (unknown)",
			(unsigned long)(pasn->Duration * wfex->nSamplesPerSec + 0.5));
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_LENGTH, buf);

/**************************************************************************
 *      Min/avg/max(total) block size:
 **************************************************************************/

	for (i = 0; (DWORD)i < pasn->dwActualLength; ++i)
	{
		unsigned long dwSize = pasn->sample_list[i].size;

		pInfo.ui64StreamBytes += dwSize;

		if (dwSize <= pInfo.dwFrameMinSize)
			pInfo.dwFrameMinSize = dwSize;

		if (dwSize >  pInfo.dwFrameMaxSize)
			pInfo.dwFrameMaxSize = dwSize;
	}

	_snprintf(buf, sizeof(buf),
		"%lu / %lu / %lu (%I64u)",
		pInfo.dwFrameMinSize,
		(unsigned long)((double)pInfo.ui64StreamBytes / (double)pasn->dwActualLength + 0.5),
		pInfo.dwFrameMaxSize,
		pInfo.ui64StreamBytes
	);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_FRAMESIZES, buf);

/**************************************************************************
 *      Data rate:
 **************************************************************************/

	if (pasn->Duration > 0)
		i = (int)((8.0 * pasn->i64TotalBytes) / pasn->Duration + 0.5);
	else
		i = 0;

	_snprintf(buf, sizeof(buf), "%i bits/sec", i);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_AUDIO_DATARATE, buf);

/**************************************************************************
 *      Language
 **************************************************************************/

	s = thisPtr->GetLanguageA(pasn->iLanguageIndex);
	if (s != NULL)
	{
		char *desc = map_language(s);
		if (desc != NULL)
		{
			_snprintf(buf, sizeof(buf), "%s (%s)", desc, s);
			delete[] desc;
		}
		else
		{
			_snprintf(buf, sizeof(buf), "Unknown (%s)", s);
		}
		delete[] s;
	}
	else
	{
		_snprintf(buf, sizeof(buf), "(not specified)");
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_LANGUAGE, buf);
}


void InputFileWMV::UpdateDlgVideo(HWND hDlg, int index)
{
	char buf[64];
	int i;

	double fps;
	char *s;

	struct {
		unsigned __int64 ui64StreamBytes;
		unsigned __int64 ui64FrameTotalSize[2];
		unsigned long dwFrameCount[2];
		unsigned long dwFrameMinSize[2];
		unsigned long dwFrameAvgSize[2];
		unsigned long dwFrameMaxSize[2];
	} pInfo = {
		0,
		{ 0, 0 },
		{ 0, 0 },
		{ 0xFFFFFFFF, 0xFFFFFFFF },
		{ 0, 0 },
		{ 0, 0 }
	};

	const ASFStreamNode    *pasn;
	const BITMAPINFOHEADER *bmih;

	InputFileWMV *thisPtr =
		(InputFileWMV *)GetWindowLongPtrA(hDlg, DWLP_USER);

	for (i = 0; i < thisPtr->num_streams; ++i)
	{
		if (thisPtr->streams[i]->fccType == streamtypeVIDEO)
		{
			if (index == 0)
			{
				pasn = thisPtr->streams[i];
				break;
			}
			--index;
		}
	}
	if (i == thisPtr->num_streams) return;

	bmih = (BITMAPINFOHEADER *)pasn->pFormat;

/**************************************************************************
 *      Frame size, average fps:
 **************************************************************************/

	if (pasn->dwScale != 0)
		fps = (double)pasn->dwRate / (double)pasn->dwScale;
	else
		fps = 0.0;

	_snprintf(buf, sizeof(buf), "%ldx%ld, %.3f fps",
		bmih->biWidth, bmih->biHeight, fps
	);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_FORMAT, buf);

/**************************************************************************
 *      Number of frames (duration):
 **************************************************************************/

	s = time_to_string(pasn->Duration);
	if (s != NULL)
	{
		_snprintf(buf, sizeof(buf), "%lu (%s)",
			pasn->dwPerceivedLength, s);

		delete[] s;
	}
	else
	{
		_snprintf(buf, sizeof(buf), "%lu (unknown)",
			pasn->dwPerceivedLength);
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_NUMFRAMES, buf);

/**************************************************************************
 *      VCM codec (fourcc):
 **************************************************************************/

	if (pasn->hCodec != NULL)
	{
		ICINFO ici;

		ici.dwSize = sizeof(ici);
		if (ICGetInfo(pasn->hCodec, &ici, sizeof(ici))
			>= FIELD_OFFSET(ICINFO, szDriver))
		{
			i = _snprintf(buf, sizeof(buf), "%ls ", ici.szDescription);
		} else {
			i = _snprintf(buf, sizeof(buf), "Unknown ");
		}
	}
	else
	{
		i = _snprintf(buf, sizeof(buf), "Missing codec ");
	}

	if (i >= 0 && i < (sizeof(buf) - 6))
	{
		char fcc[4];
		s = buf + i;

		*(DWORD *)fcc = bmih->biCompression;
		for (i = 0; i < 4; ++i)
		{
			if (fcc[i] < 32) fcc[i] = '?';
		}

		_snprintf(s, sizeof(buf) - i, "(%c%c%c%c)",
			fcc[0], fcc[1], fcc[2], fcc[3]);
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_COMPRESSION, buf);

/**************************************************************************
 *      Number of key frames:
 **************************************************************************/

	for (i = 0; (DWORD)i < pasn->dwActualLength; ++i)
	{
		const ASFSampleInfo& asi = pasn->sample_list[i];
		int iFrameType = (asi.flags & ASI_FLAG_KEYFRAME)? 0: 1;
		unsigned long dwSize = asi.size;

		pInfo.ui64StreamBytes += dwSize;

		++pInfo.dwFrameCount[iFrameType];
		pInfo.ui64FrameTotalSize[iFrameType] += dwSize;

		if (dwSize < pInfo.dwFrameMinSize[iFrameType])
			pInfo.dwFrameMinSize[iFrameType] = dwSize;

		if (dwSize > pInfo.dwFrameMaxSize[iFrameType])
			pInfo.dwFrameMaxSize[iFrameType] = dwSize;
	}

	_snprintf(buf, sizeof(buf), "%lu", pInfo.dwFrameCount[0]);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_NUMKEYFRAMES, buf);

/**************************************************************************
 *      Min/avg/max(total) key frame size:
 **************************************************************************/

	if (pInfo.dwFrameCount[0] != 0)
	{
		_snprintf(buf, sizeof(buf),
			"%lu / %lu / %lu (%I64u)",
			pInfo.dwFrameMinSize[0],
			(unsigned long)((double)pInfo.ui64FrameTotalSize[0] / (double)pInfo.dwFrameCount[0] + 0.5),
			pInfo.dwFrameMaxSize[0],
			pInfo.ui64FrameTotalSize[0]
		);
	}
	else
	{
		_snprintf(buf, sizeof(buf), "(no key frames)");
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_KEYFRAMESIZES, buf);

/**************************************************************************
 *      Min/avg/max(total) delta frame size:
 **************************************************************************/

	if (pInfo.dwFrameCount[1] != 0)
	{
		_snprintf(buf, sizeof(buf),
			"%lu / %lu / %lu (%I64u)",
			pInfo.dwFrameMinSize[1],
			(unsigned long)((double)pInfo.ui64FrameTotalSize[1] / (double)pInfo.dwFrameCount[1] + 0.5),
			pInfo.dwFrameMaxSize[1],
			pInfo.ui64FrameTotalSize[1]
		);
	}
	else
	{
		_snprintf(buf, sizeof(buf), "(no delta frames)");
	}

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_NONKEYFRAMESIZES, buf);

/**************************************************************************
 *      Data rate:
 **************************************************************************/

	if (pasn->Duration > 0)
		i = (int)((8.0 * pasn->i64TotalBytes) / pasn->Duration + 0.5);
	else
		i = 0;

	_snprintf(buf, sizeof(buf), "%i bits/sec", i);

	buf[sizeof(buf) - 1] = '\0';
	SetDlgItemTextA(hDlg, IDC_VIDEO_DATARATE, buf);
}


INT_PTR APIENTRY InputFileWMV::InfoDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	InputFileWMV *thisPtr;

	switch (message) {

		case WM_INITDIALOG:
		{
			char buf[32];
			uint32 ver;

			SetWindowLongPtrA(hDlg, DWLP_USER, lParam);
			thisPtr = (InputFileWMV *)lParam;

			ver = (*VDGetPluginInfo())->mVersion;

			_snprintf(buf, sizeof(buf), "WMV Plugin v%i.%i Information",
				(ver >> 24) & 0xFF, (ver >> 16) & 0xFF);

			buf[sizeof(buf) - 1] = '\0';
			SetWindowTextA(hDlg, buf);

			UpdateDlgVideo(hDlg, 0);

			if (thisPtr->num_audio_streams > 1)
			{
				HWND hwndSpin = GetDlgItem(hDlg, IDC_SPIN1);
				EnableWindow(hwndSpin, TRUE);
				SendMessageA(hwndSpin, UDM_SETRANGE, 0,
					(LPARAM)MAKELONG((thisPtr->num_audio_streams - 1), 0));
				SendMessageA(hwndSpin, UDM_SETPOS, 0,
					(LPARAM)MAKELONG(0, 0));
				ShowWindow(hwndSpin, SW_SHOW);
			}

			UpdateDlgAudio(hDlg, 0);

			return TRUE;
		}

		case WM_COMMAND:
			// Note: "IDCANCEL" is the [X] button
			// Return TRUE if we process the message

			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
			{
				EndDialog(hDlg, TRUE);
				return TRUE;
			}
			break;

		case WM_NOTIFY:
		{
			const NMUPDOWN *nmud = (NMUPDOWN *)lParam;

			if (nmud->hdr.idFrom == IDC_SPIN1 &&
			    nmud->hdr.code == UDN_DELTAPOS)
			{
				InputFileWMV *thisPtr = (InputFileWMV *)GetWindowLongPtrA(hDlg, DWLP_USER);
				
				int iNewPos = nmud->iPos + nmud->iDelta;

				if (iNewPos < 0)
				{
					iNewPos = thisPtr->num_audio_streams - 1;
				}
				else if (iNewPos >= thisPtr->num_audio_streams)
				{
					iNewPos = 0;
				}

				UpdateDlgAudio(hDlg, iNewPos);
			}

			break;
		}

    }

    return FALSE;
}

